/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_concurrency_model.h"
#include "nest_concurrency_pot.h"

    // private helper functions
void * workerLoop( void * );
NestReplyStatus workerRequestHandler( nestConcurrencyPot *pot,
									  int thread_id,
									  nestClient *client );

bool doDirectCallback = false;

// Derive from nestConcurrencyBase so they have a comment ancestry
// for the callback
struct ThreadData {
	// Thread specific stuff
	int					thread_id;
	pthread_t			thread;
	struct ThreadData	*next;			// Next thread

	// Give me access to everything else
	nestConcurrencyPot	*pot;
};

    // constructor
nestConcurrencyPot::nestConcurrencyPot (
        int pool_size,
		fd_set *readfds,
		fd_set *writefds,
		int id,
		nestStorageManager *sm,
		nestProtocolManager *pm,
		nestConcurrencyManager *cm )
    : nestConcurrencyModel( id )   // call base class constructor
{

    this->pendingTasks = new nestClientPool( sm, true );
    this->doneTasks    = new nestClientPool( sm, true );
	this->manager = cm;
	this->protocol_manager = pm;
	(void) writefds;

        // set up a pipe to communicate to server
    int pipe[2];
    if ( NestSpipe( pipe ) < 0 ) {
        nest_perror( "NestSpipe" );
        assert( 0 );
    }
    NestSetFileFlags( pipe[0], O_NDELAY );
    NestFd_set( pipe[0], readfds );
    maxfd = pipe[1];

    this->readfds = readfds;

        // set up the args
    this->poolSize       = pool_size;
    this->server_pipe[0] = pipe[0];
    this->server_pipe[1] = pipe[1];
    this->model_id       = id;
    nest_pthread_mutex_init( &(this->mutex), NULL );

        // pass args to each newly created thread worker
	int			tid;
	this->first_thread = NULL;
	ThreadData	*prev = NULL;
    for ( tid = 0; tid < pool_size; tid++ )
    {
		// Allocate thread specific structure
		struct ThreadData   *tdata;
		tdata = (struct ThreadData *) NestMalloc( sizeof(struct ThreadData) );
		if ( ! tdata ) {
			nest_die( "Can't allocate thread space for thread %d\n", tid );
			assert( 0 );
		}

		// Fill it in
		tdata->thread_id = tid;		// The thread's "ID"
		tdata->pot = this;			// Pointer back to myself
		tdata->next = NULL;			// Currently no next

		// Save off pointer
		if ( prev ) {
			prev->next = tdata;
		}
		prev = tdata;

		// Store off a pointer to the first one
		if ( ! this->first_thread ) {
			this->first_thread = tdata;
		}

		// And, finally, create the thread
        nest_pthread_create( &(tdata->thread),
							 NULL,
							 workerLoop,
							 (void *)tdata );
    }
}

nestConcurrencyPot::~nestConcurrencyPot( ) {
	// TODO:  This is crude, but will probably work unless the
	// threads are doing real work at the time...
	ThreadData	*tdata = this->first_thread;
	while( tdata ) {
		pthread_cancel( tdata->thread );
		tdata = tdata->next;
	}

	// Join them
	tdata = this->first_thread;
	while( tdata ) {
		ThreadData	*next;
		pthread_join( tdata->thread, NULL );

		// And, free up the memory
		next = tdata->next;
		NestFree( tdata );
		tdata = next;
	}
}

bool
nestConcurrencyPot::directCallback()
{
    return doDirectCallback;
}

char *
nestConcurrencyPot::toString()
{
    static char name[] = "pot";
    return name;
}

int
nestConcurrencyPot::getMaxfd( )
{
    return maxfd;
}

NestReplyStatus
nestConcurrencyPot::addTask( nestClient *client )
{
    NestReplyStatus status = NEST_SUCCESS;
    pendingTasks->addClient( client );
    return status;
}

NestReplyStatus
nestConcurrencyPot::makeProgress( fd_set *r_ready, fd_set *w_ready )
{
    NestReplyStatus status = NEST_SUCCESS;

        // check to see if the server has started ignoring us.
        // why would that happen?
    if ( ! NestFd_isset( server_pipe[0], readfds ) ) {
        nest_die( "Somehow the dispatcher has decided to ignore "
                  "the thread pipe.\n" );
    }

        // clear the server signal pipe here so we can signal later
    if ( NestFd_isset( server_pipe[0], r_ready ) ) {
        status = clearNestSignal();
        if ( status != NEST_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "Unable to clear pot signal: %s\n", 
					   NestErrorString( status ) );
        } 
    }

    if ( ! doDirectCallback ) {
            // call taskDone only in admin thread
        while( doneTasks->size() ) {
        //for( int i = 0; i < thread_args->doneTasks->size(); i++ ) {
            nestClient *client = doneTasks->getFront();
            assert( client != NULL );
            NestReplyStatus status;
            taskFunction *fp = client->getTaskFunction();
            if ( fp != NULL ) {
                status = client->getTaskStatus();
            } else {
                assert( client->getRequest() != NULL );
                status = client->getRequest()->getStatus();
            }
            manager->taskDone( client,
							   status,
							   model_id );
        }
    } else {
        assert( doneTasks->size() == 0 );
    }
    return status;
}

// this function is called to make sure that server breaks out of its main
// select loop after a task has been completed
// otherwise the server will not be listening to the client
void
signalNest( int pipe, int id )
{
    static char msg[] = "POT wakeup\n";
    NestWriten( pipe, msg, NestStrlen( msg ) );    
    nest_debug( NEST_DEBUG_POT, "Thread %d signaled dispatcher.\n", id );
}

// this function is called after the server breaks out of the select loop
// just clears the signal so that it can be used again
NestReplyStatus
nestConcurrencyPot::clearNestSignal( )
{
    static char msg[FILEBUFFER];
    int n = 0;
    int clear = 0;
    nest_debug( NEST_DEBUG_POT, "Clearing nest signals from pot\n" );

    nest_pthread_mutex_lock( &mutex );
    while (( n = NestRead( server_pipe[0], msg, FILEBUFFER )) > 0){
        nest_debug( NEST_DEBUG_POT, "Cleared %d bytes: %s\n", n, msg );
        clear++;
    }
    nest_pthread_mutex_unlock( &mutex );

        // if nothing was cleared return error
        // otherwise return success as long as errno is reasonable
    if ( clear == 0 ) return NEST_TEMPORARILY_UNAVAILABLE;
    return ( n == 0 || ( n == -1 && errno == EAGAIN ) )
            ? NEST_SUCCESS : ErrnoToNest( errno );
}

// this is the function where each worker thread lives
void *
workerLoop( void *void_args )
{

    struct ThreadData *data = ( struct ThreadData * ) void_args;
    NestReplyStatus status = NEST_SUCCESS;
	nestConcurrencyPot	*pot = data->pot;

    /*
    nest_debug( NEST_DEBUG_POT | NEST_DEBUG_ADMIN, 
            "Thread %d (pid %d) reporting\n", data->thread_id, NestGetpid() );
    */

        // get a request from the queue, deal with it, repeat
    nestClient *currentClient;
    while( true ) {
        currentClient = pot->pendingTasks->getFront( true );
        assert( currentClient != NULL );

        const char *request_type = ( currentClient->getRequest() == NULL )
                                 ? "special task"
                                 : currentClient->getRequest()->toString();

        nest_debug( NEST_DEBUG_POT, 
                "Thread %d handling request %s from client %s (%d)\n", 
                data->thread_id, 
                request_type, 
                currentClient->getUserIdentity(),
                currentClient->getId() );
        status = workerRequestHandler( pot, data->thread_id, currentClient );
        nest_debug( NEST_DEBUG_POT, 
                "Thread %d handled request %s from client %s (%d): %s\n", 
                data->thread_id, 
                request_type,
                currentClient->getUserIdentity(),
                currentClient->getId(), 
                NestErrorString( status ) );


            // return responsibility to the server and signal it, unlock mutex
        if ( doDirectCallback ) {
            nest_die( "I doubt doDirectCallback works - johnbent\n" );
                // lock the mutex and do shared cleanup and stats collection
            nest_pthread_mutex_lock( &(pot->mutex) );
			pot->manager->taskDone( currentClient, status, pot->model_id );
            nest_pthread_mutex_unlock( &(pot->mutex) );
        } else {
            pot->doneTasks->addClient( currentClient );
        }
        signalNest( pot->server_pipe[1], data->thread_id );
    }

}

NestReplyStatus
workerRequestHandler( nestConcurrencyPot *pot,
					  int worker_id,
					  nestClient *client )
{
    NestReplyStatus status = NEST_SUCCESS;

        // special case if client provides their own concurrency function
    taskFunction *fp = client->getTaskFunction();
    if ( fp != NULL ) {
        NestReplyStatus status = fp( client );
        nest_debug( NEST_DEBUG_PROXY, 
            "Thread %d handled special request from client %s(%d): %s\n", 
            worker_id, client->getUserIdentity(), client->getId(),
            NestErrorString( status ) );
        client->setTaskStatus( status );
        return status;
    }

    // TODO if 3pt, start it here
    // any 3pt is initiated here
    switch( client->getRequest()->getType() ) {
        case CLIENT_SEND_FILE:
                // TODO check error value
            client->sendInitialAck( );
            break;
        case CLIENT_RECV_FILE:
                // TODO check error value
            client->sendTransferInfo( );
            break;
        case THIRD_PARTY_GET:
        case THIRD_PARTY_PUT:
            nest_debug( NEST_DEBUG_POT, "Thread %d doing 3pt\n", worker_id);
            status = client->openThirdParty(
				pot->protocol_manager,
				(nestProtocolType) client->getRequest()->getRemoteProtocol() );
            break;
        case PROXY_GET:
        case PROXY_PUT:
                // nothing to do
            nest_debug( NEST_DEBUG_PROXY, "Thread %d starting proxy op.\n",
                    worker_id );
            break;
        case COPY_REQUEST:
            // nothing to do
            break; 
        default:
            assert( 0 );
            break;
    }

    if ( status != NEST_SUCCESS ) {
        client->getRequest()->setStatus( status );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

        // now go ahead and do the actual transfer 
    nest_debug( NEST_DEBUG_POT, "Thread %d enter blockingTask\n", worker_id );
    status = client->getTransfer()->BlockingTransfer( );
    nest_debug( NEST_DEBUG_POT, "Thread %d Task complete: %s\n", 
                worker_id, NestErrorString( status ) );

        // clean up after a 3pt
    if (   client->getRequest()->getType() == THIRD_PARTY_PUT
        || client->getRequest()->getType() == THIRD_PARTY_GET ) 
    {
        client->endThirdPartyTransfer( );
    }
        
    if ( status != NEST_SUCCESS ) {
        if ( status != NEST_LOT_NOT_ENOUGH ) {
            if ( errno != EPIPE ) {
                nest_warn( __FILE__, __FUNCTION__,
						   "BlockingTask failure: %s.\n",
                        NestErrorString( status ) );
            }
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
    } else {
            // now end the transfer
        status =
			( client->sendFinalAck() ? NEST_SUCCESS : NEST_NO_CONNECTION );
    }

    client->getRequest()->setStatus( status );
    return status;
}
