/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_worker_pool.h"
#include "nest_concurrency_manager.h"
#include "nest_filetransfer.h"

// NOTE that nest_worker and nest_worker_pool object speak nest_protocol
// between them, regardless of what protocol is spoken by the client

#define WORKER_ITERATOR list< nestWorker *>::iterator

nestWorkerPool::nestWorkerPool(
	int quantity,
	fd_set *watchfds,
	int (*worker_loop)( int ),
	nestProtocolManager *pm
	)
{
    maxfd = 0;
    worker_id = 0;
    this->worker_loop = worker_loop;
	this->protocol_manager = pm;
    createWorkers( quantity, watchfds );
}

int
nestWorkerPool::createWorkers( int quantity, fd_set *watchfds )
{
	int fd[2];
	pid_t pid;
	char  idstring[NEST_MESSAGESIZE];

	memset( idstring, 0, NEST_MESSAGESIZE );

	for( int i = 0; i < quantity; worker_id++, i++ ) {

		// set up stream pipe for the child
		if ( NestSpipe( fd ) < 0 ) {
			nest_perror( "s_pipe" );
			assert( 0 );
			return ERROR_EXIT;
		}

		// fork for each child
		if( ( pid = fork() ) < 0 ) {
			nest_perror( "fork" );
			assert( 0 );
			return ERROR_EXIT;
		} else if ( pid > 0) {                  /* parent */
			NestClose( fd[1] );
			addWorker( pid, fd[0], worker_id, watchfds );
		} else {                                /* child */
			// NestClose the unneeded pipe and setup the other pipes
			// arrange the other pipes as stdin and stdout for children
			NestClose( fd[0] );
			if ( fd[1] != STDIN_FILENO ) {
				NestDup2( fd[1], STDIN_FILENO );
			}

			if ( fd[1] != STDOUT_FILENO ) {
				NestDup2( fd[1], STDOUT_FILENO );
			}

			NestSnprintf( idstring, sizeof(idstring), "%d", worker_id );
			worker_loop( worker_id );

			// child should never get here
			nest_warn( __FILE__, __FUNCTION__,
					   "Weirdness in nestWorkers %d\n", __LINE__ );
		}
            
	}
    return 0;
}

int
nestWorkerPool::addWorker( pid_t pid, int stream, int id, fd_set *watchfds )
{
    nestWorker *newWorker = new nestWorker(
		pid, stream, id, protocol_manager );

    workers.push_back( newWorker );
    maxfd = ( maxfd > stream ) ? maxfd : stream;

    if ( watchfds ) NestFd_set( stream, watchfds ); 

    return workers.size();
}

int
nestWorkerPool::getMaxfds( ) {
    return maxfd;
}

ssize_t
nestWorkerPool::size( ) {
	return workers.size();
}

int
nestWorkerPool::removeWorker( pid_t pid, fd_set *watchfds, fd_set *readyfds )
{
    nestWorker *worker = getWorker( pid );

    if (worker == NULL) {
        return ERROR_EXIT;
    }

    if ( worker->getClient() != NULL ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Removing a worker %d with active client connection.\n",
				   pid);
    }

    if ( watchfds ) NestFd_clear( worker->getStream(), watchfds );
    if ( readyfds ) NestFd_clear( worker->getStream(), readyfds );
    NestClose( worker->getStream() );
        // be careful about removing the workers because we may be in a signal
        // handler and when we return execution, we might have screwed up our
        // iterator, just set the guy busy so we won't use him anymore
    //workers.remove( worker );
    //delete( worker );
    nest_debug( NEST_DEBUG_POP, "Removing dead worker %d\n", worker->getID() );
    worker->setIdle( WORKER_DEAD );
        // TODO when a worker dies while performing a transfer, we should 
        // check client connection.  If still good, server needs to listen 
        // to client again?  Otherwise, client gets stuck. 
    return 0;
}

void
nestWorkerPool::debugLevel( long level ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;
    int busy = 0, failures = 0;

    char msg[NEST_MESSAGESIZE];
    NestSnprintf( msg, NEST_MESSAGESIZE, "%d %ld\n", DEBUG_REQUEST, level );

    while( p != workers.end() ) {
        worker = *p;
        if ( ! worker->isIdle() ) {
            busy++;
        } else {
            worker->sendMessage( msg );
                // explicitly wait for ack here
            int unused;
            struct IO_Measurement unused_measure;
            if (NEST_SUCCESS != ReceiveWorkerMessage( 
					worker->getStream(), unused, unused, &unused_measure ) )
            {
                failures++;
            }
        }
        p++; 
    }
    if ( busy ) {
		nest_warn( __FILE__, __FUNCTION__,
				   "Unable to set debug in %d busy workers\n",
				   busy );
	}
    if ( failures ) {
		nest_warn( __FILE__, __FUNCTION__,
				   "Failed to set debug in %d workers\n",
				   failures );
	}
    return;
}


nestWorker*
nestWorkerPool::getWorkerFromID( int id ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;
    
    while( p != workers.end() ) {
        worker = *p;
        if ( worker->getID() == id ) return worker;
        p++;
    }
    return NULL;
}


nestWorker*
nestWorkerPool::getWorker( pid_t pid ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;
    
    while( p != workers.end() ) {
        worker = *p;
        if ( worker->getPid() == pid ) return worker;
        p++;
    }
    return NULL;
}

// getting an idle worker automatically turns idle off, so make sure to use
// the worker
// eventually, we should add timestamps to when idle was turned off and reclaim
// any slackers
// TODO: make this constant time ??
nestWorker* 
nestWorkerPool::getIdleWorker( ) {

	WORKER_ITERATOR p = workers.begin();
	nestWorker *worker;

	while( p != workers.end() ) {
		worker = *p;
		if ( worker->isIdle() ) {
			worker->setIdle( false );    
			return worker;
		}
		p++;
	}
	return NULL;
}

/* 
returns a ptr to the worker which sent the message which is written into
the msg parameter, returns NULL if there are no more worker messages
*/
nestWorker *
nestWorkerPool::canReceiveMessage( fd_set *readyfds ) {

    WORKER_ITERATOR p = workers.begin();
    nestWorker *worker;

    while( p != workers.end() ) {
        worker = *p;
        if ( worker->canReceiveMessage( readyfds ) ) {
            return worker;
        }
        p++;
    }
    return NULL;
}

// used by worker processes
bool
nestWorkerPool::SendWorkerMessage( NestConnection con, int t_id, int client_id,
        int client_socket, int status, nestFiletransfer *transfer )
{
    char msg[NEST_MESSAGESIZE];
    long bytes_transferred = 0;
    int first_in_time = 0, first_in_size = 0;
    int first_out_time = 0, first_out_size = 0;
    long total_out_time = 0, total_in_time = 0;

    if ( transfer != NULL ) {
        struct IO_Measurement *measures = transfer->getMeasurements();
        assert( measures != NULL );
        bytes_transferred = transfer->getBytesTransferred();
        first_in_time = measures->first_in_time;
        first_out_time = measures->first_out_time;
        total_in_time = measures->total_in_time;
        total_out_time = measures->total_out_time;
        first_in_size = measures->first_in_size;
        first_out_size = measures->first_out_size;
    }

    
    NestSnprintf( msg, NEST_MESSAGESIZE, 
            "%d %d %d %d %d %d %ld %d %d %ld %ld %d %d\n",
            NEST_SUCCESS, 
            t_id, 
            REQUEST_DONE, 
            status, 
            client_id,
            client_socket, 
            bytes_transferred,
            first_in_time,
            first_out_time,
            total_in_time,
            total_out_time,
            first_in_size,
            first_out_size );
    nest_debug( NEST_DEBUG_CHIRP | NEST_DEBUG_POP, 
            "W%d: Sending %s to boss.\n", t_id, msg );
    return SendNestMessage( con, msg );
}

NestReplyStatus
nestWorkerPool::ReceiveWorkerMessage( NestConnection con, 
        int &client_id, int &client_socket, struct IO_Measurement *measures )
{
    char msg[NEST_MESSAGESIZE];
    int worker_id; 
    NestReplyStatus worker_status, success, status;

        // receive message
    status = ReceiveNestMessage( con, msg );
    if ( status != NEST_SUCCESS ) return status;

        // parse message
    if ( sscanf( msg, "%d %d %d %d %d %d %ld %d %d %ld %ld %d %d",
            &status, 
            &worker_id, 
            &worker_status, 
            &success,
            &client_id, 
            &client_socket, 
            &(measures->total_bytes),
            &(measures->first_in_time), 
            &(measures->first_out_time),
            &(measures->total_in_time),
            &(measures->total_out_time), 
            &(measures->first_in_size),
            &(measures->first_out_size) ) != 13 ) 
    {
        nest_warn( __FILE__, __FUNCTION__,
				   "recvWorkerMsg: bad format msg: %s\n", msg );
        return NEST_NO_CONNECTION;
    } else {
        nest_debug( NEST_DEBUG_CHIRP | NEST_DEBUG_POP,
                "recv'd worker msg: %s\n", msg );
    }
 
        // check return values
    if ( status != NEST_SUCCESS ) return status; 
    if ( worker_status != REQUEST_DONE ) return NEST_UNKNOWN_ERROR;

    return success; 
}           
